#include <string.h>

/* 
 * main.c --
 *
 *	This file contains the main program for "wish", a windowing
 *	shell based on Tk and Tcl.  It also provides a template that
 *	can be used as the basis for main programs for other Tk
 *	applications.
 *
 * Copyright (c) 1990-1993 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#include <stdlib.h>
#include <stdio.h>
#include <iostream.h>
#include <tcl.h>
#include <tk.h>
#include <string.h>
#include "ssdbm.h"
#include "ssdbm-tcl.h"


/*
 * Declarations for various library procedures and variables (don't want
 * to include tkInt.h or tkConfig.h here, because people might copy this
 * file out of the Tk source directory to make their own modified versions).
 */

//extern void		exit _ANSI_ARGS_((int status));
//extern int		isatty _ANSI_ARGS_((int fd));
//extern int		read _ANSI_ARGS_((int fd, char *buf, size_t size));
//extern char *		strrchr _ANSI_ARGS_((CONST char *string, int c));

/*
 * Global variables used by the main program:
 */

static Tk_Window mainWindow;	/* The main window for the application.  If
				 * NULL then the application no longer
				 * exists. */
static Tcl_Interp *interp;	/* Interpreter for this application. */
char *tcl_RcFileName = NULL;	/* Name of a user-specific startup script
				 * to source if the application is being run
				 * interactively (e.g. "~/.wishrc").  Set
				 * by Tcl_AppInit.  NULL means don't source
				 * anything ever. */

static char errorExitCmd[] = "exit 1";

/*
 * Command-line options:
 */

static int synchronize = 0;
static char *fileName = NULL;
static char *name = NULL;
static char *display = NULL;
static char *geometry = NULL;
static char *mpid = NULL;
static char *mtmp = NULL;
static char *mhotlist = NULL;
static char *subject_space = NULL;
static char *icon = NULL;
static char *simon_path = NULL;
static char *help = NULL;

static Tk_ArgvInfo argTable[] = {
    {"-file", TK_ARGV_STRING, (char *) NULL, (char *) &fileName,
	"File from which to read commands"},
    {"-ss", TK_ARGV_STRING, (char *) NULL, (char *) &subject_space,
	"Subject space to open on startup"},
    {"-geometry", TK_ARGV_STRING, (char *) NULL, (char *) &geometry,
	"Initial geometry for window"},
    {"-display", TK_ARGV_STRING, (char *) NULL, (char *) &display,
	"Display to use"},
    {"-name", TK_ARGV_STRING, (char *) NULL, (char *) &name,
	"Name to use for application"},
    {"-sync", TK_ARGV_CONSTANT, (char *) 1, (char *) &synchronize,
	"Use synchronous mode for display server"},
    {"-mpid", TK_ARGV_STRING, (char *) NULL, (char *) &mpid,
	"PID of runing mosaic process"},
    {"-mhotlist", TK_ARGV_STRING, (char *) NULL, (char *) &mhotlist,
	"mosaic hotlist file"},
    {"-mtempdir", TK_ARGV_STRING, (char *) NULL, (char *) &mtmp,
	"tmp directory of runing mosaic process"},
    {"-icon", TK_ARGV_STRING, (char *) NULL, (char *) &icon,
	"optional icon to use (bitmap file)"},
    {"-path", TK_ARGV_STRING, (char *) NULL, (char *) &simon_path,
	"optional path where simon looks for it's files (makes this path home dir)"},

    {"-help", TK_ARGV_STRING, (char *) NULL, (char *) &help,
	"optional URL for help (defaults to SIMON home server)"},
    {(char *) NULL, TK_ARGV_END, (char *) NULL, (char *) NULL,
	(char *) NULL}
};


/*
 * Forward declarations for procedures defined later in this file:
 */

/*
 *----------------------------------------------------------------------
 *
 * main --
 *
 *	Main program for Wish.
 *
 * Results:
 *	None. This procedure never returns (it exits the process when
 *	it's done
 *
 * Side effects:
 *	This procedure initializes the wish world and then starts
 *	interpreting commands;  almost anything could happen, depending
 *	on the script being interpreted.
 *
 *----------------------------------------------------------------------
 */

int
main(int argc, char** argv)
{
    char *args, *p, *msg;
    char buf[20];
    int code;

    interp = Tcl_CreateInterp();
#ifdef TCL_MEM_DEBUG
    Tcl_InitMemory(interp);
#endif

     /* Set the "tcl_interactive" variable.
    Tcl_SetVar(interp, "tcl_interactive", 1, TCL_GLOBAL_ONLY);


    /*
     * Parse command-line arguments.
     */

    if (Tk_ParseArgv(interp, (Tk_Window) NULL, &argc, argv, argTable, 0)
	    != TCL_OK) {
	fprintf(stderr, "%s\n", interp->result);
	exit(1);
    }

    // set application name
    if (name == NULL) {
	if (fileName != NULL) {
	    p = fileName;
	} else {
	    p = argv[0];
	}
	name = strrchr(p, '/');
	if (name != NULL) {
	    name++;
	} else {
	    name = p;
	}
    }
	

    /*
     * If a display was specified, put it into the DISPLAY
     * environment variable so that it will be available for
     * any sub-processes created by us.
     */

    if (display != NULL) {
	Tcl_SetVar2(interp, "env", "DISPLAY", display, TCL_GLOBAL_ONLY);
    }

    /*
     * Initialize the Tk application.
     */

    mainWindow = Tk_CreateMainWindow(interp, display, name, "Tk");
    if (mainWindow == NULL) {
	fprintf(stderr, "%s\n", interp->result);
	exit(1);
    }
    if (synchronize) {
	XSynchronize(Tk_Display(mainWindow), True);
    }
    Tk_GeometryRequest(mainWindow, 200, 200);

    /*
     * Make command-line arguments available in the Tcl variables "argc"
     * and "argv".  Also set the "geometry" variable from the geometry
     * specified on the command line.
     */

    args = Tcl_Merge(argc-1, argv+1);
    Tcl_SetVar(interp, "argv", args, TCL_GLOBAL_ONLY);
    ckfree(args);
    sprintf(buf, "%d", argc-1);
    Tcl_SetVar(interp, "argc", buf, TCL_GLOBAL_ONLY);
    Tcl_SetVar(interp, "argv0", (fileName != NULL) ? fileName : argv[0],
	    TCL_GLOBAL_ONLY);
    if (geometry != NULL) {
	Tcl_SetVar(interp, "geometry", geometry, TCL_GLOBAL_ONLY);
    }


    // set tcl variables

    if (mpid!=NULL)
	Tcl_SetVar (interp, "mpid", mpid, TCL_GLOBAL_ONLY);
      else Tcl_SetVar (interp, "mpid", "", TCL_GLOBAL_ONLY);
    if (mtmp!=NULL)
	Tcl_SetVar (interp, "mtmp", mtmp, TCL_GLOBAL_ONLY);
      else Tcl_SetVar (interp, "mtmp", "/tmp", TCL_GLOBAL_ONLY);
    if (mhotlist!=NULL)
	Tcl_SetVar (interp, "mhotlist", mhotlist, TCL_GLOBAL_ONLY);
      else Tcl_SetVar (interp, "mhotlist", "~/.mosaic-hotlist-default", 
							TCL_GLOBAL_ONLY);
    if (help!=NULL)
	Tcl_SetVar (interp, "help", help, TCL_GLOBAL_ONLY);
      else Tcl_SetVar (interp, "help",
		 "http://web.elec.qmw.ac.uk:12121/simon.help.html", 
							TCL_GLOBAL_ONLY);
    if (icon!=NULL)
	Tcl_SetVar (interp, "icon", icon, TCL_GLOBAL_ONLY);
      else Tcl_SetVar (interp, "icon", "null", TCL_GLOBAL_ONLY);
    if (simon_path!=NULL)
	Tcl_SetVar (interp, "simon_path", simon_path, TCL_GLOBAL_ONLY);
      else Tcl_SetVar (interp, "simon_path", "", TCL_GLOBAL_ONLY);
    if (subject_space!=NULL)
	Tcl_SetVar (interp, "subject_space", subject_space, TCL_GLOBAL_ONLY);
      else Tcl_SetVar (interp, "subject_space", "", TCL_GLOBAL_ONLY);

    /*
     * Invoke application-specific initialization.
     */

    if (Tcl_AppInit(interp) != TCL_OK) {
	fprintf(stderr, "Tcl_AppInit failed: %s\n", interp->result);
    }


    /*
     * Invoke the script specified on the command line
     */

    if (fileName != NULL) {
	String tcl_file;
	if (simon_path!=NULL) {tcl_file+=simon_path; tcl_file+="/";};
	tcl_file+=fileName;
	code = Tcl_VarEval(interp, "source ", (char*)tcl_file, (char *) NULL);
	if (code != TCL_OK) {
	    cout<<"error in, or can't find .tcl script\n"; exit (-1) ;
	}
    }
     else {cout << "I must have a -file arg, eg -file simple-simon.tcl\n";
								 exit (-1);};


    /*
     * Set the geometry of the main window, if requested.
     */

    if (geometry != NULL) {
	code = Tcl_VarEval(interp, "wm geometry . ", geometry, (char *) NULL);
	if (code != TCL_OK) {
	    fprintf(stderr, "%s\n", interp->result);
	}
    }

 // create a subject space database manager

    if (subject_space == NULL) 
      	{
	cout << "I must have a -ss arg, eg -ss .simon_ss_default\n";
	exit (-1);
	};
    SS_Error error;
    SSDBM ssdbm(error, subject_space);
    if (error!=OK) 
	{cout<<"error opening subject space\n"; exit(-1);};
    ssdbm_ptr=&ssdbm;


    fflush(stdout);

    /*
     * Loop infinitely, waiting for commands to execute.  When there
     * are no windows left, Tk_MainLoop returns and we exit.
     */

    Tk_MainLoop();

    /*
     * Don't exit directly, but rather invoke the Tcl "exit" command.
     * This gives the application the opportunity to redefine "exit"
     * to do additional cleanup.
     */

    Tcl_Eval(interp, "exit");
    exit(1);

}



/*
 *----------------------------------------------------------------------
 *
 * Tcl_AppInit --
 *
 *	This procedure performs application-specific initialization.
 *	Most applications, especially those that incorporate additional
 *	packages, will have their own version of this procedure.
 *
 * Results:
 *	Returns a standard Tcl completion code, and leaves an error
 *	message in interp->result if an error occurs.
 *
 * Side effects:
 *	Depends on the startup script.
 *
 *----------------------------------------------------------------------
 */

int
Tcl_AppInit(Tcl_Interp *interp)
{
    Tk_Window main;

    main = Tk_MainWindow(interp);

    /*
     * Call the init procedures for included packages.  Each call should
     * look like this:
     *
     * if (Mod_Init(interp) == TCL_ERROR) {
     *     return TCL_ERROR;
     * }
     *
     * where "Mod" is the name of the module.
     */

    if (Tcl_Init(interp) == TCL_ERROR) {
	return TCL_ERROR;
    }
    if (Tk_Init(interp) == TCL_ERROR) {
	return TCL_ERROR;
    }

    /*
     * Call Tcl_CreateCommand for application-specific commands, if
     * they weren't already created by the init procedures called above.
     */

    /*
     * Specify a user-specific startup file to invoke if the application
     * is run interactively.  Typically the startup file is "~/.apprc"
     * where "app" is the name of the application.  If this line is deleted
     * then no user-specific startup file will be run under any conditions.
     */

    register_new_tcl_commands(interp);

   // tcl_RcFileName = "simonrc";
    return TCL_OK;
}
