
#include "ssdbm.h"
#include <BitSet.h>
#include "functions.h"


SS_Error SSDBM::add_resource_to_subjects (const String& url, 
			 Set_String_& subject_ids)
   {
   SS_Error error;
   Record original_rec, new_rec;
   error = load_record(url, original_rec);
   if (error==OK)
	{
	new_rec=original_rec;
	new_rec.subject_ids+=subject_ids;
	error = update_subjects_after_additions_to_record
					(original_rec, new_rec, PASSIVE);
	if (error==OK) save_record (new_rec, GDBM_REPLACE);
	};
   return error;
   };

SS_Error SSDBM::remove_resource_from_subjects (const String& url, 
				      		Set_String_& subject_ids)
   {
   SS_Error error;
   Record original_rec, new_rec;
   error = load_record(url, original_rec);
   if (error==OK)
	{
	new_rec=original_rec;
	new_rec.subject_ids-=subject_ids;
	if (new_rec.subject_ids.no_of_members()!=0)
	  {
	  error = update_subjects_after_deletions_from_record
	  				(original_rec, new_rec, PASSIVE);
	  if (error==OK) save_record (new_rec, GDBM_REPLACE);
	  }
	 else error=ILEGAL_REQUEST;
	};
   return error;
   };



SS_Error SSDBM::new_resource (const Resource& resource, 
					Set_String_& subject_ids)
   {
//cout << "in SSDBM::new_resource\n";
   SS_Error error;
   Record empty_rec;
   Record new_rec (resource,subject_ids);
   error = update_subjects_after_additions_to_record
						(empty_rec, new_rec, PASSIVE);
   if (error==OK) error=save_record (new_rec, GDBM_INSERT);
   return error;
   };


SS_Error SSDBM::update_resource (const Resource& resource)
   {
   Record rec;
   SS_Error error = load_record(resource.url, rec);
   if (error==OK)
	{
	rec.resource=resource;
	error = save_record(rec, GDBM_REPLACE);
	};
   return error;
   };


SS_Error SSDBM::new_subject (const String& subject_id,
					const String& parent_id)
   {
   SS_Error error;
   Subject sub_parent;
   Subject sub;
   sub.subject_id=subject_id;
 
   if (parent_id.length()!=0)
      {
      error = load_subject(parent_id, sub_parent);
      sub.parent_subjects.add_member(parent_id);
      sub_parent.child_subjects.add_member(subject_id);
      }
    else error = OK;
   if (error == OK)
     {
     error = save_subject(sub, GDBM_INSERT);
     if ((error==OK)&&(parent_id.length()!=0))
	{
  	error = save_subject(sub_parent, GDBM_REPLACE);
	};
     };
   return error;
   };

SS_Error SSDBM::add_subject
			 (const String& subject_id, const String& parent_id)
   {
   SS_Error error;
   Subject sub_parent;
   Subject sub;
   error = load_subject(subject_id, sub);
   if (error==OK)
     {
     error = load_subject(parent_id, sub_parent);
     if (error==OK)
	{
	sub.parent_subjects.add_member(parent_id);
	sub_parent.child_subjects.add_member(subject_id);
	error = save_subject(sub, GDBM_REPLACE);
	if (error==OK) error=save_subject(sub_parent, GDBM_REPLACE);
	};
     };
   return error;
   };	 

SS_Error SSDBM::add_subject
		(const String& subject_id,  /*psconst*/ Set_String_& parent_ids)
   {
   SS_Error error, temp_error;
   error=OK;
   parent_ids.next_member_returns_first();
   while (parent_ids.can_use_get_next_member())
	{
	temp_error=add_subject(subject_id, parent_ids.get_next_member());
	if (temp_error!=OK) error=temp_error;
	};
   return error;
   };	 

SS_Error SSDBM::remove_subject
			 (const String& subject_id,const String& child_id)
   {
   SS_Error error;
   Subject sub_child;
   Subject sub;
   error = load_subject(subject_id, sub);
   if (error==OK)
     {
     error = load_subject(child_id, sub_child);
     if (error==OK)
	{
	sub.parent_subjects.delete_member(child_id);
	sub_child.parent_subjects.delete_member(subject_id);
	error = save_subject(sub, GDBM_REPLACE);
	if (error==OK) error=save_subject(sub_child, GDBM_REPLACE);
	};
     };
   return error;
   };	 

     
SS_Error SSDBM::delete_resource_from_all_subjects
				(const String& url)
   {
   Record rec_original, rec_new;
   SS_Error error = load_record(url, rec_original);
   if (error==OK)
	{
	rec_new.resource.url=rec_original.resource.url;
	//ie rec_new has the same name as rec_original, but it's in no subjects
	error=update_subjects_after_deletions_from_record
					(rec_original, rec_new, ACTIVE);
	if (error==OK)
	  {
          key_datum.dptr=url;
	  key_datum.dsize=url.length()+1;
	  gdbm_delete(records_file, key_datum);
	  };
	}; 
   return error;
   };


SS_Error SSDBM::subjects_of_resource
			 (const String& url, Set_String_& subject_ids)
   {
   SS_Error error;
   Record record;
   error=load_record(url, record);
   if (error==OK) subject_ids=record.subject_ids;
   return error;
   };


SS_Error SSDBM::set_subjects_of_resource
			 (const String& url, Set_String_& subject_ids)
 //this function sets the subjects of the resource, reguardless of the resources
 //previous subjects, all updates are ACTIVE.
   {
   SS_Error error;
   Record original_rec, intermediate_rec, new_rec;
   error = load_record(url, original_rec);
   if (error==OK)
	{
	intermediate_rec=original_rec;
	intermediate_rec.subject_ids+=subject_ids;
	error = update_subjects_after_additions_to_record
				  (original_rec, intermediate_rec, ACTIVE);
        if (error==OK) 
	   {
	   new_rec=original_rec;
	   new_rec.subject_ids=subject_ids;
	   error = update_subjects_after_deletions_from_record
				(intermediate_rec, new_rec, ACTIVE);
 	   if (error==OK) save_record (new_rec, GDBM_REPLACE);
	   };
	};
   return error;
   };

SS_Error SSDBM::urls_in_intersection 
			(/*psconst*/ Set_String_& subject_ids,
			 Set_String_& urls)
   {
   SS_Error error;
   Subject temp_sub;
   subject_ids.next_member_returns_first();
   for (int n=1; n<=subject_ids.no_of_members(); n++)
	{
	error = load_subject(subject_ids.get_next_member(), temp_sub);
	if (error!=OK) break;
	if (n==1) urls=temp_sub.urls;
	  else    urls*=temp_sub.urls;
	};
   return error;
   };

SS_Error SSDBM::resources_in_intersection 
	     (/*psconst*/ Set_String_& subject_ids, Set_Resource_& resources)
   {
   SS_Error error;
   Set_String_ urls;
   Record rec;
   error = urls_in_intersection(subject_ids, urls);
   if (error==OK)
     {
     urls.next_member_returns_first();
     while (urls.can_use_get_next_member())
	{
	error = load_record(urls.get_next_member(), rec);
	if (error!=OK) break;
	resources.add_member(rec.resource);
	};
     };
   return error;
   };

SS_Error SSDBM::parents_of_subject 
			(const String& subject_id, Set_String_& parent_ids)
   {
   SS_Error error;
   Subject sub;
   error=load_subject(subject_id,sub);
   if (error==OK) parent_ids=sub.parent_subjects;
   return error;
   };

SS_Error SSDBM::children_of_subject 
			(const String& subject_id, Set_String_& child_ids)
   {
   SS_Error error;
   Subject sub;
   error=load_subject(subject_id,sub);
   if (error==OK) child_ids=sub.child_subjects;
   return error;
   };


SS_Error SSDBM::intersectors_of_subject 
			(const String& subject_id, Set_String_& intersector_ids)
   {
   SS_Error error;
   Subject sub;
   error=load_subject(subject_id,sub);
   if (error==OK)
	{
	// now must get all the subjects of all the records of sub
	Record temp_rec;
	sub.urls.next_member_returns_first();
	while (sub.urls.can_use_get_next_member())
	  {
	  error=load_record(sub.urls.get_next_member(),temp_rec);
	  if (error==OK)
	    intersector_ids+=temp_rec.subject_ids;
	   else break;
	  };
	};
   return error;
   };

SS_Error SSDBM::delete_subject (const String& subject_id)
 //NB this operation is ileagal if it leaves subjectless records
   {
   SS_Error error;
   Subject original;
   Subject empty_original;
   error = load_subject(subject_id, original);
   if (error==OK)
     {
     empty_original.subject_id=subject_id;
     error = update_records_after_deletions_from_subject
				(original, empty_original, PASSIVE);	
     if (error==OK)
	{
	// temp needed here, otherwise simplified() would return a tempary ob.
	String temp = simplified(subject_id);
	key_datum.dptr=temp;
	key_datum.dsize=temp.length()+1;
	gdbm_delete(subjects_file, key_datum);
	
	//now make all children excommunicate this parent
	Subject child;
	original.child_subjects.next_member_returns_first();
	while (original.child_subjects.can_use_get_next_member())
	  {
	  error = load_subject
			   (original.child_subjects.get_next_member(), child);
	  if (error!=OK) break;
	  child.parent_subjects.delete_member(subject_id);
	  error = save_subject (child, GDBM_REPLACE);	  
	  if (error!=OK) break;
	  };
	};
     };
   return error;
   };


SS_Error SSDBM::update_subjects_after_additions_to_record
		  (const Record& original_record, const Record& new_record,
		   const SS_Update_Mode mode)
 // must make sure that i) new subjects exist, if mode ACTIVE make new
								// subjects
 //		       ii) new subjects in record made to reference record
   {
   Set_String_ new_subject_ids = new_record.subject_ids;
   new_subject_ids-=original_record.subject_ids;
   SS_Error error=OK;
   Subject* subjects=new Subject[new_subject_ids.no_of_members()];
   new_subject_ids.next_member_returns_first();
   String temp_subject_id;
   int n;
   for (n=0; n<new_subject_ids.no_of_members(); n++)
	{
	temp_subject_id=new_subject_ids.get_next_member();
	error = load_subject(temp_subject_id, subjects[n]);
	if (error==OK) 
 	  subjects[n].urls.add_member(new_record.resource.url); 
	 else
	   if ((mode==ACTIVE)&&(error==SUBJECT_NOT_FOUND)) 
		{
		error = new_subject (temp_subject_id, "");
		subjects[n].subject_id=temp_subject_id;
	 	subjects[n].urls.add_member(new_record.resource.url); 
		};
	if (error!=OK) break;
	};
   if (error==OK)
	for (int n=0; n<new_subject_ids.no_of_members(); n++)
		save_subject(subjects[n], GDBM_REPLACE);
   delete[] subjects;
   return error;
   };

SS_Error SSDBM::update_subjects_after_deletions_from_record
		  (const Record& original_record, const Record& new_record,
						  const SS_Update_Mode mode)
 // must make sure that references are removed from the subjects
 // if mode ACTIVE, chuck away subjects that are empty & unrelated
   {
   Set_String_ stale_references = original_record.subject_ids;
	       stale_references -= new_record.subject_ids;
   Subject sub;
   SS_Error error=OK;
   stale_references.next_member_returns_first();
   while (stale_references.can_use_get_next_member())
	{
	error = load_subject(stale_references.get_next_member(), sub);
	if (error == OK)
		{
		sub.urls.delete_member(original_record.resource.url);
		error = save_subject(sub, GDBM_REPLACE);
		}
	    else break;
	};
   if ((error==OK)&&(mode==ACTIVE))
	{
	stale_references.next_member_returns_first();
	while (stale_references.can_use_get_next_member())
	  {
	  error = load_subject(stale_references.get_next_member(), sub);
	  if (error == OK)
	    	if ( 	(sub.urls.no_of_members()==0) &&
			(sub.parent_subjects.no_of_members()==0) &&
			(sub.child_subjects.no_of_members()==0) )
		  error = delete_subject (sub.subject_id);
	  if (error!=OK) break;
	  }
        }
   return error;
   };


SS_Error SSDBM::update_records_after_additions_to_subject
		  (const Subject& original_subject, const Subject& new_subject,
						    const SS_Update_Mode mode)
 //must add references to the new subject, in all the resource records that
 //have been added to new_subject
   {
 //Set_String_ new_references = ((new_subject.urls)-(original_subject.urls));
   Set_String_ new_references=new_subject.urls;
		new_references-=original_subject.urls; 

   // now add the new references
   // NB we don't assume that all the new references exist
   SS_Error error=OK;
   //Record records[new_references.no_of_members()];
   Record* records= new Record[new_references.no_of_members()*sizeof(Record)];
   int n;
   new_references.next_member_returns_first();
   for (n=0; n<new_references.no_of_members(); n++)
	{
	error=load_record (new_references.get_next_member(), records[n]);
	if (error==OK)
	   records[n].subject_ids.add_member(original_subject.subject_id);
	 else break;
	};
   if (error==OK)
	for (n=0; n<new_references.no_of_members(); n++)
	   save_record (records[n], GDBM_REPLACE); 
   delete[] records;
   return error;
   };

SS_Error SSDBM::update_records_after_deletions_from_subject 
		  (const Subject& original_subject, const Subject& new_subject,
						      SS_Update_Mode mode)
 //must remove references to original_subject from all records that have been 
 //removed from original_subject
   {
   SS_Error error = OK;
   int deletion_makes_a_subjectless_record=0;
 //Set_String_ stale_references = (original_subject.urls)-(new_subject.urls);
   Set_String_ stale_references=original_subject.urls;
		stale_references-=new_subject.urls;

   // get rid of stale references
   if (mode=PASSIVE)  // ie must not end up with records that are in
     { 			    					// no subjects
     int n;
     Record records[stale_references.no_of_members()];
     stale_references.next_member_returns_first();
     for (n=0; n<stale_references.no_of_members(); n++)
	{
	error=load_record (stale_references.get_next_member(), records[n]);
	if (error!=OK) break; 
	records[n].subject_ids.delete_member(original_subject.subject_id);
	if (records[n].subject_ids.no_of_members()==0) 
		deletion_makes_a_subjectless_record=1;
	if (deletion_makes_a_subjectless_record==1) break;
	};
     if ((deletion_makes_a_subjectless_record==1)&&(error==OK))
	 error=ILEGAL_REQUEST;
      else
	{
	for (n=0; n<stale_references.no_of_members(); n++)
	  	{
		error=save_record (records[n], GDBM_REPLACE);
		if (error!=OK) break;
		};
	};
     }
    else   // mode == ACTIVE, ie can delete records that end up in no 
     {								// subjects
     Record rec;
     stale_references.next_member_returns_first();
     for (int n=0; n<stale_references.no_of_members(); n++)
	{
	load_record (stale_references.get_next_member(), rec);
	rec.subject_ids.delete_member(original_subject.subject_id);
	if (rec.subject_ids.no_of_members()==0)
		{
		key_datum.dptr=rec.resource.url;
		key_datum.dsize=rec.resource.url.length()+1;
		gdbm_delete(records_file, key_datum);
		}
	  else  save_record (rec, GDBM_REPLACE);
	};
      };  
    
   return error;
   };


SS_Error SSDBM::save_subject (Subject& subject, int flag)
   {
   SSByteStore data (100,80);
   SSByteStore key (20,4);
   
   key.store(simplified(subject.subject_id));
   key_datum.dptr = (char*)key.pointer();
   key_datum.dsize = key.used_size();

   data.store(subject);

   data_datum.dptr = (char*)data.pointer();
   data_datum.dsize = data.used_size();

   int ret =  (gdbm_store(subjects_file, key_datum, data_datum, flag));
   if (ret==0) return OK;
   if (ret==1) return SUBJECT_ALREADY_EXISTS;
   else return FATAL_SERVER_ERROR; 

}; //end ::save_subject


SS_Error SSDBM::load_subject (const String& subject_id, 
				     Subject& subject)
   {
   SS_Error ret=OK;
   SSByteStore data (0,80);
   SSByteStore key (20,4);

   key.store(simplified(subject_id));
   key_datum.dptr = (char*) key.pointer();
   key_datum.dsize = key.used_size();

   data_datum = gdbm_fetch (subjects_file, key_datum);
   if (data_datum.dptr==NULL)   
	{
	ret=SUBJECT_NOT_FOUND;
	}
     else
	{
	data.take_over(data_datum.dptr, data_datum.dsize);
	data.get(subject);
	};
   return ret;    
   
}; //end ::load_subject



SS_Error  SSDBM::exists_resource (const String& url)
   {
   SS_Error ret;
   SSByteStore key (20,4);

   key.store(url);
   key_datum.dptr = (char*) key.pointer();
   key_datum.dsize = key.used_size();
   
   data_datum = gdbm_fetch (records_file, key_datum);
   if (data_datum.dptr==NULL)   
	ret=RESOURCE_NOT_FOUND;
     else
	ret=RESOURCE_ALREADY_EXISTS;
   return ret;
   }; 

SS_Error SSDBM::save_record (Record& record, int flag)
   {
   SSByteStore data (100,80);
   SSByteStore key (20,4);
   
   key.store(record.resource.url);
   key_datum.dptr=(char*)key.pointer();
   key_datum.dsize=key.used_size();
      
   data.store(record);
  
   data_datum.dptr=(char*)data.pointer();
   data_datum.dsize=data.used_size();

   int ret = (gdbm_store (records_file, key_datum, data_datum, flag));
   if (ret==0) return OK;
   if (ret==1) return RESOURCE_ALREADY_EXISTS;
   else return FATAL_SERVER_ERROR; 
   };

SS_Error SSDBM::get_resource (const String& url, Resource& resource)
   {
   SS_Error error;
   Record rec;
   error = load_record (url, rec);
   if (error == OK) resource=rec.resource;
   return error;
   };


SS_Error SSDBM::load_record (const String& url, Record& record)
   {
   SS_Error ret=OK;
   SSByteStore data (0,80);
   SSByteStore key (20,4);

   key.store(url);
   key_datum.dptr = (char*) key.pointer();
   key_datum.dsize = key.used_size();
   
   data_datum = gdbm_fetch (records_file, key_datum);
   if (data_datum.dptr==NULL)   
	{
	ret=RESOURCE_NOT_FOUND;
	}
     else
	{
	record.resource.url = url;
	data.take_over(data_datum.dptr, data_datum.dsize);
	data.get(record);
	};

   return ret;
   }; 


SS_Error SSDBM::urls_list (Set_String_& urls)
   { 
   datum temp_key;
   key_datum=gdbm_firstkey(records_file);
   while (key_datum.dptr!=NULL)
	{
	urls.add_member(key_datum.dptr, NO_TEST_FOR_DUPLICATE);
	temp_key=key_datum;
	key_datum=gdbm_nextkey(records_file,temp_key);
	free (temp_key.dptr);
	};
   return OK;
   };


SS_Error SSDBM::subjects_list (Set_String_& subjects)
   {
   datum temp_key;
   key_datum=gdbm_firstkey(subjects_file);
   while (key_datum.dptr!=NULL)
	{
	subjects.add_member(key_datum.dptr, NO_TEST_FOR_DUPLICATE);
	temp_key=key_datum;
	key_datum=gdbm_nextkey(subjects_file,temp_key);
	free (temp_key.dptr);
	};
   return OK;
   };
   

void SSDBM::reorganise()
   {
   gdbm_reorganize(records_file);
   gdbm_reorganize(subjects_file);
   };


void SSDBM::text_dump (String& dump)
 // fill dump with entire subject space contents
 // format is ssci commands that reproduce the subject space, ie
 //	newsubject <subject_id>
 //	addsubject <child_id> <parent_id>
 //	newresource <url> <name> <subject> <subject> ...
 // section separator is |= at start of an otherwise empty line
 // 
   {
   Set_String_ subjects;
   Set_String_ urls;
   subjects_list(subjects);
   urls_list(urls);

   subjects.next_member_returns_first();
   while (subjects.can_use_get_next_member())
	{dump+= "newsubject "; 
	 dump+="\""; dump+=subjects.get_next_member(); dump+="\"\n";
	};
   
   String tmp;
   Set_String_ parents;
   subjects.next_member_returns_first();
   while (subjects.can_use_get_next_member())
	{
	subjects.get_next_member(tmp);
	parents_of_subject(tmp,parents);
	if (parents.no_of_members()!=0) 
	  {
	  dump+="addsubject "; 
	  dump+="\""; dump+=tmp; dump+="\" ";
	  parents.next_member_returns_first();
	  while (parents.can_use_get_next_member())
	  	{dump+="\""; dump+=parents.get_next_member(); dump+="\" ";};
	  dump += '\n';
	  };
	};	
   
   Record record;
   urls.next_member_returns_first();
   while (urls.can_use_get_next_member())
	{
	load_record(urls.get_next_member(), record);
	dump+="newresource ";
	dump+="\""; dump+=record.resource.url; dump+="\" ";
	dump+="\""; dump+=record.resource.name; dump+="\" ";
	record.subject_ids.next_member_returns_first();
	while(record.subject_ids.can_use_get_next_member())
	 {dump+="\""; dump+=record.subject_ids.get_next_member(); dump+="\" ";};
	dump+="\n";
	};

   };


SSDBM::SSDBM (SS_Error& status, const String& name, 
				      int flags=GDBM_WRCREAT)
   {
   String temp;
   temp = name+record_file_extention;
   records_file=gdbm_open (temp.chars(),
		gdbm_file_block_size, flags, gdbm_file_mode, 0);
   //NB IGNOR THESE WARNINGS ON LINE ABOVE & BELOW
   //#warning ignore warnings around this line
   temp = name+subject_file_extention;
   subjects_file=gdbm_open (temp.chars(),
		gdbm_file_block_size, flags, gdbm_file_mode, 0);

   if ((records_file!=NULL)&&(subjects_file!=NULL))
	  status=OK;
     else status=DATABASE_NOT_FOUND;
   }; //end ::SSDBM




SSDBM::~SSDBM ()
   {
   //gdbm_sync(records_file);
   //gdbm_sync(subjects_file);
   if ((records_file!=NULL)&&(subjects_file!=NULL)) 
	   reorganise();
   if (records_file!=NULL)
	   gdbm_close(records_file);  
   if (subjects_file!=NULL)
	   gdbm_close(subjects_file);
   }; //end ::~subject_database


SS_Error SS_Map_Manager::resources_in_node
		(Node_id node_id, Set_Resource_& resources)
   {
   SS_Error error;
   Set_Resource_ temp_res;
   Set_String_ S;
   error = intersectors_of_node(node_id, S);
   S-=node_id;
   if (error==OK)
      error = resources_in_intersection(node_id, resources);
   while ( (error==OK)&&(S.can_use_get_next_member()) )
	{
   	Set_String_ temp_subs;
	temp_subs.add_member(S.get_next_member());
	error = resources_in_intersection(temp_subs, temp_res);
	resources-=temp_res;
	};
   if (error==OK) if (resources.no_of_members()==0) error = NO_SUCH_NODE;
   return error;
   };

SS_Error SS_Map_Manager::intersectors_of_node
	(Node_id node_id, Set_String_& intersectors)
   {
   SS_Error error;
   Set_String_ temp;
   error = intersectors_of_subject(node_id.get_first_member(), intersectors);
   if (error!=OK) return error;
   while (node_id.can_use_get_next_member())
	{	
	error = intersectors_of_subject(node_id.get_next_member(), temp);
	if (error!=OK) return error;
	intersectors*=temp;
	};
   return OK;
   };

SS_Error SS_Map_Manager::subjects_close_to_node
	(Node_id node_id, Set_String_& close_subjects)
   {
cout<<"\nsubjects_close_to_node\n";
   SS_Error error;
   Set_String_ temp;
   error = intersectors_of_subject(node_id.get_first_member(), close_subjects);
   if (error!=OK) return error;
   while (node_id.can_use_get_next_member())
	{	
	error = intersectors_of_subject(node_id.get_next_member(), temp);
	if (error!=OK) return error;
	close_subjects+=temp;
	};
cout<<"\nsubjects_close_to_node return\n";
   return OK;
   };
  	
SS_Error SS_Map_Manager::possible_unit_distance_nodes
		   (Node_id node_id, Set_Node_id_& possible_unit_distance_nodes)
   {
   SS_Error error;
   //find specialised_nodes
   Set_String_ subject_ids_for_adding_to_node_id;
   Set_String_ intersectors;
   error = intersectors_of_node
		(node_id, intersectors);
   if (error!=OK) return error;
   subject_ids_for_adding_to_node_id = intersectors;
   subject_ids_for_adding_to_node_id -= node_id;
   subject_ids_for_adding_to_node_id.next_member_returns_first();
   while (subject_ids_for_adding_to_node_id.can_use_get_next_member())
	{
	Node_id temp = node_id; 
	temp.add_member(subject_ids_for_adding_to_node_id.get_next_member());
	possible_unit_distance_nodes.add_member(temp);
	};
   //find generalised_nodes
   Set_String_ subject_ids_for_removing_from_node_id = node_id;
   while (subject_ids_for_removing_from_node_id.can_use_get_next_member())
	{
	Node_id temp = node_id;
	temp.delete_member
		(subject_ids_for_removing_from_node_id.get_next_member());
	possible_unit_distance_nodes.add_member(temp);
	};	
   return error;
   };



SS_Error SS_Map_Manager::neighbor_nodes
  			 (Node_id node_id, Set_Node_id_& neighbor_nodes)
   {
   SS_Error error;
   Set_String_ subject_ids;
  // error=subjects_close_to_node(node_id, subject_ids);
   error=intersectors_of_node(node_id, subject_ids);
   if (error!=OK) return error;

   for (int distance=1; ( distance<=2 )
			&&(neighbor_nodes.no_of_members()<=10);
	    distance++)
	{
	Set_Node_id_ possible_node_ids;
	generate_node_id_combinations(node_id, subject_ids, distance, 
						    possible_node_ids);
   	possible_node_ids.next_member_returns_first();
   	while (possible_node_ids.can_use_get_next_member())
	  {
	  Node_id possible_node_id=possible_node_ids.get_next_member();
	  Set_Resource_ dummy;
	  error = resources_in_node (possible_node_id, dummy);
	  if (error==OK)
		  neighbor_nodes.add_member(possible_node_id);
	   else if (error==NO_SUCH_NODE)
		  error = OK;
                 else /*return error*/ error=OK;
	  };
	};
   return error;
   };

void SS_Map_Manager::generate_node_id_combinations
		(Node_id& node_id, Set_String_& subject_ids, int distance,
					   Set_Node_id_& generated_node_ids)
 // Take the subjects in subject_ids, and use them as to generate combinations
 // of subjects that define node_ids that are the specified distance away from 
 // node_id.  Does not check for null node_ids.

   {
   int i;  // i will be used in loops to move through the index

 //make index array of subject_ids;
   int last_index=subject_ids.no_of_members();
   String index[last_index+1];        		//first index is 0
   subject_ids.next_member_returns_first();
   for (i=1; i<=last_index; i++)
	index[i]=subject_ids.get_next_member();

 //make bitset from node_id;   
   BitSet node_bs;
   node_bs.clear();
   for (i=0; i<=last_index; i++)
	if (node_id.has_member(index[i]))
		node_bs.set(i);

 //generate cobination vectors
   int no_of_generators = nCr(subject_ids.no_of_members(),distance);
   BitSet generators[no_of_generators];
   generate_bit_combinations(subject_ids.no_of_members(),distance,1,generators);
 //generate node_ids
   for (i=0; i<no_of_generators; i++)
     {
     BitSet result_bs;
     Node_id result_id;
     xor(node_bs, generators[i], result_bs);
     for (int b=1; b<=subject_ids.no_of_members(); b++)
	if (result_bs.test(b)) result_id.add_member(index[b]);
     generated_node_ids.add_member(result_id);
     };
   };


void SS_Map_Manager::generate_bit_combinations
			(unsigned u, unsigned d, unsigned s, 
						BitSet* result)
	// u = bitset size
	// d = bits to place
	// s = start placing this bit
	// result = ptr to BitSet=nCr(u,d)] in effect,
	// call by reference, done this way so caller can use stack space 
	// for array
   {
   unsigned result_counter=0;
   BitSet mask1;
   // inside the for loop, is a recursive call
   if (d>0)  // test for stopping condition of recursive loop
    {
    for (unsigned p=s; (p+d-1<=u); p++)
	{
	mask1.clear();
	mask1.set(p);
	unsigned  mask2s_size;
	mask2s_size = nCr(u-p,d-1);
	BitSet mask2s[mask2s_size];
	generate_bit_combinations(u, d-1, p+1, mask2s);
	for (unsigned mask2s_counter=0; mask2s_counter<mask2s_size; 
							mask2s_counter++)
	   xor(mask2s[mask2s_counter],mask1,
			result[result_counter++]);
	};
   
    };   
   };


