#ifndef _ssdbm_h
#define _ssdbm_h 1


#include <unistd.h>
#include <iostream.h>
#include <stdlib.h>
#include <gdbm.h>   /* gnu database managmer */
#include "ss.h"
#include "functions.h"
#include <errno.h>
#include <BitSet.h>


const String subject_file_extention = ".subject";
const String record_file_extention = ".record";
const int gdbm_file_block_size = 1024;
const int gdbm_file_mode = 00664;



enum SS_Error {	OK,
		FATAL_SERVER_ERROR,
		BAD_REPLY_FILE,
		UNRECOGNISED_REQUEST,
		DATABASE_NOT_FOUND,
		RESOURCE_NOT_FOUND, 
		SUBJECT_NOT_FOUND,
		ILEGAL_REQUEST,
		RESOURCE_ALREADY_EXISTS,
		SUBJECT_ALREADY_EXISTS,
		NO_SUCH_NODE
	      };

enum SS_Update_Mode { ACTIVE, PASSIVE };
 



class SSDBM
  {
   public:
	// status set to 0 for OK, else -1
	SSDBM (SS_Error& status, const String& name,
			   int flags = GDBM_WRCREAT);
	//NB see gbdm.h for values of flag
	~SSDBM ();
	void get_errors(int& gdbm_errors_, int& file_errors_)
		{gdbm_errors_=gdbm_errno; file_errors_=errno;};

	// all these functions return negative error values, or 0 = OK
	// NB /*psconst*/ means psuedo const, set members are unchanged, but
	/* Set object is changed (as its iterator is used), so it can't be
	   const */
	
	SS_Error new_resource (const Resource& resource,
	       /*psconst*/ Set_String_& subject_ids);
	SS_Error update_resource (const Resource& resource);
	SS_Error delete_resource_from_all_subjects (const String& url);
	SS_Error get_resource (const String& url, Resource& resource);

	SS_Error new_subject (const String& subject_id,
			 const String& parent_id);
	SS_Error add_subject (const String& subject_id,
			  const String& parent_id);
	SS_Error add_subject (const String& subject_id,
			 /*psconst*/ Set_String_& parent_ids);
	SS_Error delete_subject (const String& subject_id);
	SS_Error remove_subject
		(const String& subject_id,const String& child_id);

	SS_Error add_resource_to_subjects (const String& url, 
			   Set_String_& subject_ids);
	SS_Error remove_resource_from_subjects (const String& url,
					 /*psconst*/ Set_String_& subject_ids);
	SS_Error urls_in_intersection 
			(/*psconst*/ Set_String_& subject_ids,
			 Set_String_& urls);
	SS_Error subjects_of_resource
			(const String& url, Set_String_& subject_ids);
	SS_Error set_subjects_of_resource
			(const String& url, Set_String_& subject_ids);
	SS_Error resources_in_intersection 
			(/*psconst*/ Set_String_& subject_ids,
			 Set_Resource_& resources);
	SS_Error parents_of_subject (const String& subject_id,
			        Set_String_& parent_ids);
	SS_Error children_of_subject (const String& subject_id,
				 Set_String_& child_ids);
	SS_Error intersectors_of_subject (const String& subject_id,
				    Set_String_& intersector_ids);
	SS_Error urls_list (Set_String_& urls);
	SS_Error subjects_list (Set_String_& subjects);
	SS_Error exists_resource (const String& url);

	void text_dump (String& dump);	
	void reorganise();
	
   protected:
	GDBM_FILE records_file;
	GDBM_FILE subjects_file;
	datum data_datum;
	datum key_datum;
	SS_Error load_record (const String& url, Record& record);	
	SS_Error load_subject (const String& subject_id, Subject& subject);
	SS_Error save_record (/*psconst*/ Record& record, int flag);
	SS_Error save_subject (/*psconst*/ Subject& subject, int flag);
	SS_Error update_records_after_additions_to_subject 
		  (const Subject& original_subject, const Subject& new_subject,
		   const SS_Update_Mode mode);
	SS_Error update_records_after_deletions_from_subject 
		  (const Subject& original_subject, const Subject& new_subject,
				  		    const SS_Update_Mode mode);
	SS_Error update_subjects_after_additions_to_record 
		  (const Record& original_record, const Record& new_record,
						  const SS_Update_Mode mode);
	SS_Error update_subjects_after_deletions_from_record
		  (const Record& original_record, const Record& new_record,
							SS_Update_Mode mode);

 	SS_Error tidy();
   };


class SS_Map_Manager : public SSDBM
   {
   public:
	SS_Map_Manager::SS_Map_Manager (SS_Error& status, const String& name, 
				                   int flags=GDBM_WRCREAT)
			 : SSDBM(status, name, flags) {};

	SS_Error resources_in_node
		   (Node_id node_id, Set_Resource_& resources);
	SS_Error neighbor_nodes
		   (Node_id node_id, Set_Node_id_& nearby_nodes);
	SS_Error possible_unit_distance_nodes
		   (Node_id node_id, Set_Node_id_& unit_distance_nodes);

	SS_Error intersectors_of_node
		(Node_id node_id, Set_String_& intersectors);
	SS_Error subjects_close_to_node
		(Node_id node_id, Set_String_& close_subjects);

	void test()
		{
		Set_String_ a;
		a.add_member (" pig ");
		a.add_member (" cow ");
		a.add_member (" hen ");
		a.add_member (" horse ");
		a.add_member (" bull ");
		a.add_member (" sheep ");
		a.add_member (" dog ");
		a.add_member (" goose ");
		Node_id b;
		b.add_member (" hen ");
		b.add_member (" horse ");
		b.add_member (" bull ");

		Set_Node_id_ c;
		generate_node_id_combinations
			(b,a,0,c);
		cout <<"\nresults 0: \n";
		c.screen_dump();

		Set_Node_id_ d;
		generate_node_id_combinations
			(b,a,1,d);
		cout <<"\nresults 1: \n";
		d.screen_dump();

		Set_Node_id_ e;
		generate_node_id_combinations
			(b,a,2,e);
		cout <<"\nresults 2: \n";
		e.screen_dump();

		Set_Node_id_ f;
		generate_node_id_combinations
			(b,a,3,f);
		cout <<"\nresults 3: \n";
		f.screen_dump();

		Set_Node_id_ g;
		generate_node_id_combinations
			(b,a,4,g);
		cout <<"\nresults 4: \n";
		g.screen_dump();



		};

   protected:
	void generate_node_id_combinations
		(Node_id& node_id, Set_String_& subject_ids, int distance,
					   Set_Node_id_& generated_node_ids);
 	void generate_bit_combinations
			(unsigned u, unsigned d, unsigned s, 
						BitSet* result);
   };





#endif
